local addonName, addonTable = ...; -- Pulls back the Addon-Local Variables and store them locally.
local addon = _G[addonName];
local L = LibStub("AceLocale-3.0"):GetLocale(addonName);

--- Localize Vars
local Enemies = addon.Modules.Enemies;
local Objects = addon.Core.Objects;
-- Objects

local Player = addon.Units.Player;
local Target = addon.Units.Target;
local Racial, Spell, Talent, Aura, Azerite, Item, Consumable;
local Variables = {};

-- Rotation Variables
local nameAPL = "lunaeclipse_druid_balance";

-- Base APL Class
local function APL(rotationName, rotationDescription, specID)
	-- Inherits APL Class so get the base class.
	local self = addonTable.rotationsAPL(rotationName, rotationDescription, specID);

	-- Store the information for the script.
	self.scriptInfo = {
		SpecializationID = self.SpecID,
		ScriptAuthor = "LunaEclipse",
		GuideAuthor = "Tettles and SimCraft",
		GuideLink = "http://www.wowhead.com/balance-druid-guide",
		WoWVersion = 80001,
	};

	self.Defensives = {
		Barkskin = function()
			return Player.DamagePredicted(6) >= 30;
		end,

		Renewal = function()
			return Player.Health.Percent() < 70
			   and Player.DamagePredicted(5) >= 25;
		end,
	};

	-- Table to hold requirements to use spells and items listed in the rotation.
	self.Requirements = {
		-- actions+=/berserking,if=buff.celestial_alignment.up|buff.incarnation.up
		Berserking = function()
			return Player.Buff(Aura.CelestialAlignment).Up()
			    or Player.Buff(Aura.IncarnationChosenOfElune).Up();
		end,

		-- actions+=/blood_fury,if=buff.celestial_alignment.up|buff.incarnation.up
		BloodFury = function()
			return Player.Buff(Aura.CelestialAlignment).Up()
			    or Player.Buff(Aura.IncarnationChosenOfElune).Up();
		end,

		-- We can't do raid add events so just skip these.
		-- actions+=/celestial_alignment,if=astral_power>=40&(!azerite.lively_spirit.enabled|buff.lively_spirit.up)&(((raid_event.adds.duration%15)*(4)<(raid_event.adds.in%180))|(raid_event.adds.up))
		CelestialAlignment = function()
			return Player.AstralPower() >= 40
			   and (not Azerite.LivelySpirit.Enabled() or Player.Buff(Aura.LivelySpirit).Up());
		end,

		-- actions+=/force_of_nature,if=(buff.celestial_alignment.up|buff.incarnation.up)|(cooldown.celestial_alignment.remains>30|cooldown.incarnation.remains>30)
		ForceOfNature = function()
			return (Player.Buff(Aura.CelestialAlignment).Up() or Player.Buff(Aura.IncarnationChosenOfElune).Up())
			    or (Spell.CelestialAlignment.Cooldown.Remains() > 30 or Talent.IncarnationChosenOfElune.Cooldown.Remains() > 30);
		end,

		-- actions+=/full_moon,if=astral_power.deficit>40+execute_time%1.5
		FullMoon = function()
			return Player.AstralPower.Deficit() > 40 + Talent.FullMoon.ExecuteTime() / 1.5;
		end,

		-- We can't do raid add events so just skip these.
		-- actions+=/fury_of_elune,if=(((raid_event.adds.duration%8)*(4)<(raid_event.adds.in%60))|(raid_event.adds.up))&((buff.celestial_alignment.up|buff.incarnation.up)|(cooldown.celestial_alignment.remains>30|cooldown.incarnation.remains>30))
		FuryOfElune = function()
			return (Player.Buff(Aura.CelestialAlignment).Up() or Player.Buff(Aura.IncarnationChosenOfElune).Up())
			    or (Spell.CelestialAlignment.Cooldown.Remains() > 30 or Talent.IncarnationChosenOfElune.Cooldown.Remains() > 30);
		end,

		-- actions+=/half_moon,if=astral_power.deficit>20+execute_time%1.5
		HalfMoon = function()
			return Player.AstralPower.Deficit() > 20 + Talent.HalfMoon.ExecuteTime() / 1.5;
		end,

		-- We can't do raid add events so just skip these.
		-- actions+=/incarnation,if=astral_power>=40&(((raid_event.adds.duration%30)*(4)<(raid_event.adds.in%180))|(raid_event.adds.up))
		IncarnationChosenOfElune = function()
			return Player.AstralPower() >= 40;
		end,

		-- We can't do raid add events so just skip these.
		-- actions+=/innervate,if=azerite.lively_spirit.enabled&(cooldown.incarnation.up|cooldown.celestial_alignment.remains<12)&(((raid_event.adds.duration%15)*(4)<(raid_event.adds.in%180))|(raid_event.adds.up))
		Innervate = function()
			return Azerite.LivelySpirit.Enabled()
			   and (Talent.IncarnationChosenOfElune.Cooldown.Up() or Spell.CelestialAlignment.Cooldown.Remains() < 12);
		end,

		LunarStrike = {
			-- actions+=/lunar_strike,if=astral_power.deficit>=16&(buff.lunar_empowerment.stack=3|(spell_targets<3&astral_power>=40&(buff.lunar_empowerment.stack=2&buff.solar_empowerment.stack=2)))&!(spell_targets.moonfire>=2&variable.az_potm=3&active_enemies=2)
			Empowerment = function(numEnemies)
				return Player.AstralPower.Deficit() >= 16
				   and (Player.Buff(Aura.LunarEmpowerment).Stack() == 3 or (Enemies.GetEnemies(Spell.LunarStrike) < 3 and Player.AstralPower() >= 40 and (Player.Buff(Aura.LunarEmpowerment).Stack() == 2 and Player.Buff(Aura.SolarEmpowerment).Stack() == 2)))
				   and not (Enemies.GetEnemies(Spell.Moonfire) >= 2 and Variables.az_potm == 3 and numEnemies == 2);
			end,

			-- actions+=/lunar_strike,if=((buff.warrior_of_elune.up|buff.lunar_empowerment.up|spell_targets>=3&!buff.solar_empowerment.up)&(!buff.celestial_alignment.up&!buff.incarnation.up|variable.az_streak<2|!prev_gcd.1.lunar_strike)|(variable.az_ds&!buff.dawning_sun.up))&!(spell_targets.moonfire>=2&active_enemies<=4&(variable.az_potm=3|variable.az_potm=2&active_enemies=2))
			Use = function(numEnemies)
				return (Player.Buff(Aura.WarriorOfElune).Up() or Player.Buff(Aura.LunarEmpowerment).Up() or Enemies.GetEnemies(Spell.LunarStrike) >= 3 and not Player.Buff(Aura.SolarEmpowerment).Up())
				   and (not Player.Buff(Aura.CelestialAlignment).Up() and not Player.Buff(Aura.IncarnationChosenOfElune).Up() or Variables.az_streak < 2 or not Player.PrevGCD(1, Spell.LunarStrike) or (Variables.az_ds > 0 and not Player.Buff(Aura.DawningSun).Up()))
				   and not (Enemies.GetEnemies(Spell.Moonfire) >= 2 and numEnemies <= 4 and (Variables.az_potm == 3 or Variables.az_potm == 2 and numEnemies == 2));
			end,
		},

		-- actions+=/moonfire,target_if=refreshable,if=astral_power.deficit>=7&target.time_to_die>6.6&(!buff.celestial_alignment.up&!buff.incarnation.up|!variable.az_streak|!prev_gcd.1.moonfire)&(movement.distance>0|raid_event.movement.in>remains|remains<=execute_time*2)
		Moonfire = function(_, Target)
			return Target.Debuff(Aura.Moonfire).Refreshable()
			   and Player.AstralPower.Deficit() >= 7
			   and Target.TimeToDie() > 6.6
			   and (not Player.Buff(Aura.CelestialAlignment).Up() and not Player.Buff(Aura.IncarnationChosenOfElune).Up() or not Variables.az_streak or not Player.PrevGCD(1, Spell.Moonfire));
		end,

		-- SimCraft doesn't mention it, but make sure we are not already in Moonkin Form before suggesting it.
		-- actions.precombat+=/moonkin_form
		MoonkinForm = function()
			return not Player.Buff(Aura.MoonkinForm).Up();
		end,

		-- actions+=/new_moon,if=astral_power.deficit>10+execute_time%1.5
		NewMoon = function()
			return Player.AstralPower.Deficit() > 10 + Talent.NewMoon.ExecuteTime() / 1.5;
		end,

		-- actions=potion,if=buff.celestial_alignment.up|buff.incarnation.up
		RisingDeath = function()
			return Player.Buff(Aura.CelestialAlignment).Up()
			    or Player.Buff(Aura.IncarnationChosenOfElune).Up();
		end,

		SolarWrath = {
			-- az_streak is a number, LUA can't do a boolean check with a number so use greater than 0 instead.
			-- actions+=/solar_wrath,if=astral_power.deficit>=12&(buff.solar_empowerment.stack=3|(variable.az_sb>1&spell_targets.starfall<3&astral_power>=32&!buff.sunblaze.up))&!(spell_targets.moonfire>=2&active_enemies<=4&variable.az_potm=3)|(variable.az_streak&(buff.celestial_alignment.up|buff.incarnation.up)&!prev_gcd.1.solar_wrath&astral_power.deficit>=12)
			Empowerment = function(numEnemies)
				return Player.AstralPower.Deficit() >= 12
				   and (Player.Buff(Aura.SolarEmpowerment).Stack() == 3 or (Variables.az_sb > 1 and Enemies.GetEnemies(40) < 3 and Player.AstralPower() >= 32 and not Player.Buff(Aura.Sunblaze).Up()))
				   and not (Enemies.GetEnemies(Spell.Moonfire) >= 2 and numEnemies <= 4 and Variables.az_potm == 3)
	                or (Variables.az_streak > 0 and (Player.Buff(Aura.CelestialAlignment).Up() or Player.Buff(Aura.IncarnationChosenOfElune).Up()) and not Player.PrevGCD(1, Spell.SolarWrath) and Player.AstralPower.Deficit() >= 12);
			end,

			-- actions+=/solar_wrath,if=(!buff.celestial_alignment.up&!buff.incarnation.up|variable.az_streak<2|!prev_gcd.1.solar_wrath)&!(spell_targets.moonfire>=2&active_enemies<=4&(variable.az_potm=3|variable.az_potm=2&active_enemies=2))
			Use = function(numEnemies)
				return (not Player.Buff(Aura.CelestialAlignment).Up() and not Player.Buff(Aura.IncarnationChosenOfElune).Up() or Variables.az_streak < 2 or not Player.PrevGCD(1, Spell.SolarWrath))
				   and not (Enemies.GetEnemies(Spell.Moonfire) >= 2 and numEnemies <= 4 and (Variables.az_potm == 3 or Variables.az_potm == 2 and numEnemies == 2));
			end,
		},

		-- actions+=/starfall,if=spell_targets.starfall>=3&(!buff.starlord.up|buff.starlord.remains>=4)
		Starfall = function(numEnemies)
			return numEnemies >= 3
			   and (not Player.Buff(Aura.Starlord).Up() or Player.Buff(Aura.Starlord).Remains() >= 4);
		end,

		-- We can't do raid movement events, so lets just skip this.
		-- actions+=/starsurge,if=(spell_targets.starfall<3&(!buff.starlord.up|buff.starlord.remains>=4)|execute_time*(astral_power%40)>target.time_to_die)&(!buff.celestial_alignment.up&!buff.incarnation.up|variable.az_streak<2|!prev_gcd.1.starsurge)&(raid_event.movement.in>(buff.lunar_empowerment.stack*action.lunar_strike.execute_time+buff.solar_empowerment.stack*action.solar_wrath.execute_time)|(astral_power+buff.lunar_empowerment.stack*12+buff.solar_empowerment.stack*8)>=96)
		Starsurge = function(numEnemies)
			return (numEnemies < 3 and (not Player.Buff(Aura.Starlord).Up() or Player.Buff(Aura.Starlord).Remains() >= 4) or Spell.Starsurge.ExecuteTime() * (Player.AstralPower() / 40) > Target.TimeToDie())
			   and (not Player.Buff(Aura.CelestialAlignment).Up() and not Player.Buff(Aura.IncarnationChosenOfElune).Up() or Variables.az_streak < 2 or not Player.PrevGCD(1, Spell.Starsurge))
			   and Player.AstralPower() + Player.Buff(Aura.LunarEmpowerment).Stack() * 12 + Player.Buff(Aura.SolarEmpowerment).Stack() * 8 > 96;
		end,

		-- actions+=/stellar_flare,target_if=refreshable,if=astral_power.deficit>=12&target.time_to_die>7.2&(!buff.celestial_alignment.up&!buff.incarnation.up|!variable.az_streak|!prev_gcd.1.stellar_flare)
		StellarFlare = function(_, Target)
			return Target.Debuff(Aura.StellarFlare).Refreshable()
			   and Player.AstralPower.Deficit() >= 12
			   and Target.TimeToDie() > 7.2
			   and (not Player.Buff(Aura.CelestialAlignment).Up() and not Player.Buff(Aura.IncarnationChosenOfElune).Up() or not Variables.az_streak or not Player.PrevGCD(1, Talent.StellarFlare));
		end,

		Sunfire = {
			-- We can't do raid movement events, so just skip these, az_streak is a number, LUA can't do a boolean check with a number so use equals 0 instead.
			-- actions+=/sunfire,target_if=refreshable,if=astral_power.deficit>=7&target.time_to_die>5.4&(!buff.celestial_alignment.up&!buff.incarnation.up|!variable.az_streak|!prev_gcd.1.sunfire)&(movement.distance>0|raid_event.movement.in>remains|remains<=execute_time*2)
			Refresh = function(_, Target)
				return Target.Debuff(Aura.Sunfire).Refreshable()
				   and Player.AstralPower.Deficit() >= 7
				   and Target.TimeToDie() > 5.4
				   and (not Player.Buff(Aura.CelestialAlignment).Up() and not Player.Buff(Aura.IncarnationChosenOfElune).Up() or Variables.az_streak == 0 or not Player.PrevGCD(1, Spell.Sunfire));
			end,

			-- az_streak is a number, LUA can't do a boolean check with a number so use equals 0 instead.
			-- actions+=/sunfire,if=(!buff.celestial_alignment.up&!buff.incarnation.up|!variable.az_streak|!prev_gcd.1.sunfire)&!(variable.az_potm>=2&spell_targets.moonfire>=2)
			Use = function(numEnemies)
				return (not Player.Buff(Aura.CelestialAlignment).Up() and not Player.Buff(Aura.IncarnationChosenOfElune).Up() or Variables.az_streak == 0 or not Player.PrevGCD(1, Spell.Sunfire))
				   and (Variables.az_potm >= 2 and numEnemies >= 2);
			end,
		},
	};

	Objects.FinalizeRequirements(self.Defensives, self.Requirements);

	-- Function for setting up action objects such as spells, buffs, debuffs and items, called when the rotation becomes the active rotation.
	function self.Enable(...)
		Racial, Spell, Talent, Aura, Azerite = ...;

		Aura.DawningSun = Objects.newSpell(276154);
		Aura.LivelySpirit = Objects.newSpell(279648);
		Aura.Sunblaze = Objects.newSpell(274399);

		Item = {};

		Consumable = {
			-- Potions
			RisingDeath = Objects.newItem(152559),
		};

		Objects.FinalizeActions(Racial, Spell, Talent, Aura, Azerite, Item, Consumable);
	end

	-- Function for setting up the configuration screen, called when rotation becomes the active rotation.
	function self.SetupConfiguration(config, options)
		config.RacialOptions(options, Racial.Berserking, Racial.Shadowmeld);
		config.AOEOptions(options, Spell.Starfall);
		config.BuffOptions(options, Spell.MoonkinForm);
		config.CooldownOptions(options, Spell.CelestialAlignment, Talent.ForceOfNature, Talent.FuryOfElune, Talent.IncarnationChosenOfElune, Talent.NewMoon, Talent.StellarFlare, Talent.WarriorOfElune);
		config.DefensiveOptions(options, Spell.Barkskin, Talent.Renewal);
	end

	-- Function for destroying action objects such as spells, buffs, debuffs and items, called when the rotation is no longer the active rotation.
	function self.Disable()
		local coreTables = addon.Core.Tables;

		coreTables.Wipe(Racial, Spell, Talent, Aura, Azerite, Item, Consumable);
	end

	-- Function for checking the rotation that displays on the Defensives icon.
	function self.Defensive(action)
		-- The abilities here should be listed from highest damage required to suggest to lowest,
		-- Specific damage types before all damage types.

		-- Protects against all types of damage
		action.EvaluateDefensiveAction(Spell.Barkskin, self.Defensives.Barkskin);

		-- Self Healing goes at the end and is only suggested if a major cooldown is not needed.
		action.EvaluateDefensiveAction(Talent.Renewal, self.Defensives.Renewal);
	end

	-- Function for displaying opening rotation.
	function self.Opener(action)
	end

	function self.PrecombatVariables(action)
		-- # Cast sw after every non-sw spell unless capping, sw sf sw if forced
		-- actions.precombat+=/variable,name=az_streak,value=azerite.streaking_stars.rank
		Variables.az_streak = Azerite.StreakingStars.Rank();
		-- actions.precombat+=/variable,name=az_ds,value=azerite.dawning_sun.rank
		Variables.az_ds = Azerite.DawningSun.Rank();
		-- actions.precombat+=/variable,name=az_sb,value=azerite.sunblaze.rank
		Variables.az_sb = Azerite.Sunblaze.Rank();
		-- actions.precombat+=/variable,name=az_potm,value=azerite.power_of_the_moon.rank,if=talent.twin_moons.enabled
		Variables.az_potm = Talent.TwinMoons.Enabled() and Azerite.PowerOfTheMoon.Rank() or 0;
	end

	-- Function for displaying any actions before combat starts.
	function self.Precombat(action)
		action.EvaluateAction(Spell.MoonkinForm, self.Requirements.MoonkinForm);
		-- actions.precombat+=/potion
		action.EvaluateAction(Consumable.RisingDeath, true);
		-- actions.precombat+=/solar_wrath
		action.EvaluateAction(Spell.SolarWrath, true);
	end

	-- Function for checking the rotation that displays on the Single Target, AOE, Off GCD and CD icons.
	function self.Combat(action)
		action.EvaluateAction(Consumable.RisingDeath, self.Requirements.RisingDeath);
		action.EvaluateAction(Racial.Berserking, self.Requirements.Berserking)

		-- # CDs
		-- actions+=/warrior_of_elune
		action.EvaluateAction(Talent.WarriorOfElune, true);
		action.EvaluateAction(Spell.Innervate, self.Requirements.Innervate);
		action.EvaluateAction(Talent.IncarnationChosenOfElune, self.Requirements.IncarnationChosenOfElune);
		action.EvaluateAction(Spell.CelestialAlignment, self.Requirements.CelestialAlignment);
		action.EvaluateAction(Talent.FuryOfElune, self.Requirements.FuryOfElune);
		action.EvaluateAction(Talent.ForceOfNature, self.Requirements.ForceOfNature);

		-- # DoTs
		action.EvaluateCycleAction(Spell.Sunfire, self.Requirements.Sunfire.Refresh);
		action.EvaluateCycleAction(Spell.Moonfire, self.Requirements.Moonfire);
		action.EvaluateCycleAction(Talent.StellarFlare, self.Requirements.StellarFlare);

		-- # Empowerment cap check
		action.EvaluateAction(Spell.LunarStrike, self.Requirements.LunarStrike.Empowerment);
		action.EvaluateAction(Spell.SolarWrath, self.Requirements.SolarWrath.Empowerment);

		-- # Rotation
		action.EvaluateAction(Spell.Starsurge, self.Requirements.Starsurge, Enemies.GetEnemies(40));
		action.EvaluateAction(Spell.Starfall, self.Requirements.Starfall, Enemies.GetEnemies(40));
		action.EvaluateAction(Talent.NewMoon, self.Requirements.NewMoon);
		action.EvaluateAction(Talent.HalfMoon, self.Requirements.HalfMoon);
		action.EvaluateAction(Talent.FullMoon, self.Requirements.FullMoon);
		action.EvaluateAction(Spell.LunarStrike, self.Requirements.LunarStrike.Use);
		action.EvaluateAction(Spell.SolarWrath, self.Requirements.SolarWrath.Use);
		action.EvaluateAction(Spell.Sunfire, self.Requirements.Sunfire.Use, Enemies.GetEnemies(Spell.Moonfire));
		-- actions+=/moonfire
		action.EvaluateAction(Spell.Moonfire, true);
	end

	return self;
end

local rotationAPL = APL(nameAPL, "LunaEclipse: Balance Druid", addonTable.Enum.SpecID.DRUID_BALANCE);