-- Pulls back the Addon-Local Variables and store them locally.
local addonName, addonTable = ...;
local addon = _G["LunaEclipse_Rotation"];

-- Plugin information variables.
local pluginName = "Game Tweaks";

-- Check whether the plugin is supposed to be active
if addon and addon.System.Plugins.IsActive(pluginName) then
	-- Store local copies of Blizzard API and other addon global functions and variables.
	local apiUnit, Blizzard, systemCallbacks, systemTimers = addon.API.Unit, addon.API.Blizzard, addon.System.Callbacks, addon.System.Timers;
	local collectgarbage = collectgarbage;

	local MemoryCleanup = {};

	MemoryCleanup = {
		Clean = function()
			if not apiUnit.InCombat() then
				collectgarbage("collect");
			end
		end,

		Use = function()
			return addon.Core.Tables.GetValue(addonTable.gameTweaks, "useMemoryCleaner", 1) == 1;
		end,
	};

	---------------------- Public Functions ----------------------
	addon.Features.MemoryCleanup = {
		-- useMemoryCleaner
		Clean = function()
			if MemoryCleanup.Use() and not apiUnit.InCombat() then
				addon.System.Events.CombatEventDelay(MemoryCleanup.Clean);
			end
		end,

		Start = function()
			if not apiUnit.InCombat() then
				local cleanupRate = addon.Core.Tables.GetValue(addonTable.gameTweaks, "cleanupRate", 30);

				-- Clean the memory and then set the timer to clean on interval
				MemoryCleanup.Clean();
				systemTimers.TimerAdd(MemoryCleanup.Clean, cleanupRate);
			end
		end,

		Stop = function()
			-- Stop timer because we don't want to clean memory anymore
			systemTimers.TimerRemove(MemoryCleanup.Clean);
		end,
	};

	----------------------- Plugin Code Operation ----------------
	local moduleEvents = {
		enterCombat = function()
			if MemoryCleanup.Use() then
				-- Stop timer because we don't want to clean in combat
				systemTimers.TimerRemove(MemoryCleanup.Clean);
			end
		end,

		leaveCombat = function()
			if MemoryCleanup.Use() then
				local cleanupRate = addon.Core.Tables.GetValue(addonTable.gameTweaks, "cleanupRate", 30);

				-- Clean the memory and then set the timer to clean on interval
				MemoryCleanup.Clean();
				systemTimers.TimerAdd(MemoryCleanup.Clean, cleanupRate);
			end
		end,
	};

	-- These are only registered if the plugin is active as they control how what is called
	-- from LunaEclipse Rotation Helper
	systemCallbacks.RegisterAddonCallbacks(moduleEvents);
end