-- Pulls back the Addon-Local Variables and store them locally.
local _, addonTable = ...;
local addon = _G["LunaEclipse_Rotation"];

-- Plugin information variables.
local pluginName = "Interrupt Manager";

-- Check whether the plugin is supposed to be active
if addon and addon.System.Plugins.IsActive(pluginName) then
	-- Store local copy addon global functions and variables
	local apiSpell, coreObjects, coreSettings, coreTables, modulesActionBar, systemCallbacks, systemEvents, systemTimers = addon.API.Spell, addon.Core.Objects, addon.Core.Settings, addon.Core.Tables, addon.Modules.ActionBar, addon.System.Callbacks, addon.System.Events, addon.System.Timers;
	local error, format, setmetatable, sort, tinsert, type, wipe = error, format, setmetatable, sort, tinsert, type, wipe;

	-- Objects
	local Pet, Player, Target = addon.Units.Pet, addon.Units.Player, addon.Units.Target;
	local spellObjects = {
		CC = {
			BlastWave = coreObjects.newSpell(157981),
			Blind = coreObjects.newSpell(2094),
			BlindingLight = coreObjects.newSpell(115750),
			BlindingSleet = coreObjects.newSpell(207167),
			FreezeTrap = coreObjects.newSpell(187650),
			Gouge = coreObjects.newSpell(1776),
			Hex = coreObjects.newSpell(51514),
			Hibernate = coreObjects.newSpell(2637),
			Imprison = coreObjects.newSpell(217832),
			IncapacitatingRoar = coreObjects.newSpell(99),
			IntimidatingRoar = coreObjects.newSpell(236748),
			IntimidatingShout = coreObjects.newSpell(5246),
			Paralysis = coreObjects.newSpell(115078),
			Polymorph = coreObjects.newSpell(118),
			PsychicScream = coreObjects.newSpell(8122),
			Repentance = coreObjects.newSpell(20066),
			ShackleUndead = coreObjects.newSpell(9484),
			SigilOfMisery = coreObjects.newSpell(207684),
			Supernova = coreObjects.newSpell(157980),
			Typhoon = coreObjects.newSpell(132469),
		},
		Interrupt = {
			CounterShot = coreObjects.newSpell(147362),
			Counterspell = coreObjects.newSpell(2139),
			Disrupt = coreObjects.newSpell(183752),
			Kick = coreObjects.newSpell(1766),
			MindFreeze = coreObjects.newSpell(47528),
			Muzzle = coreObjects.newSpell(187707),
			Pummel = coreObjects.newSpell(6552),
			Rebuke = coreObjects.newSpell(96231),
			SigilOfSilence = coreObjects.newSpell(202137),
			Silence = coreObjects.newSpell(15487),
			SkullBash = coreObjects.newSpell(106839),
			SolarBeam = coreObjects.newSpell(78675),
			SpearHandStrike = coreObjects.newSpell(116705),
			SpellLock = coreObjects.newSpell(19647),
			WindShear = coreObjects.newSpell(57994),
		},
		Stun = {
			Asphyxiate = coreObjects.newSpell(108194),
			AxeToss = coreObjects.newSpell(89766),
			BetweenTheEyes = coreObjects.newSpell(199804),
			BullRush = coreObjects.newSpell(255723),
			CapacitorTotem = coreObjects.newSpell(192058),
			ChaosNova = coreObjects.newSpell(179057),
			CheapShot = coreObjects.newSpell(1833),
			FelEruption = coreObjects.newSpell(211881),
			Gnaw = coreObjects.newSpell(47481),
			HammerOfJustice = coreObjects.newSpell(853),
			Intimidation = coreObjects.newSpell(19577),
			KidneyShot = coreObjects.newSpell(408),
			LegSweep = coreObjects.newSpell(119381),
			Maim = coreObjects.newSpell(22570),
			MightyBash = coreObjects.newSpell(5211),
			QuakingPalm = coreObjects.newSpell(107079),
			PsychicHorror = coreObjects.newSpell(64044),
			Shadowfury = coreObjects.newSpell(30283),
			Shockwave = coreObjects.newSpell(46968),
			StormBolt = coreObjects.newSpell(107570),
			WarStomp = coreObjects.newSpell(20549),
		},
	};

	local stopped = false;
	local rotationIndex = 0;
	local actions = {};
	local evaluateActions = {};

	local Events, Requirements, Suggest = {}, {}, {};

	local function sortActions(a, b)
		if coreTables.GetValue(a, "WaitTime", 0) == coreTables.GetValue(b, "WaitTime", 0) then
			return coreTables.GetValue(a, "RotationIndex", 0) < coreTables.GetValue(b, "RotationIndex", 0);
		end

		return coreTables.GetValue(a, "WaitTime") < coreTables.GetValue(b, "WaitTime");
	end

	Requirements = {
		BetweenTheEyes = function()
			return not Player.IsStealthed()
			   and Player.ComboPoints() >= 1;
		end,

		KidneyShot = function()
			return not Player.IsStealthed()
					and Player.ComboPoints() >= 1;
		end,

		Maim = function()
			return not Player.IsStealthed()
			   and Player.ComboPoints() >= 1;
		end,
	};
	coreObjects.FinalizeRequirements(Requirements);

	Suggest = {
		CrowdControl = function(action)
			local interruptManager = addonTable.interruptManager;

			if coreTables.GetValue(interruptManager, "useCC", 0) == 1 then
				action.EvaluateInterruptAction(spellObjects.CC.Imprison, true);
				action.EvaluateInterruptAction(spellObjects.CC.Hibernate, true);
				action.EvaluateInterruptAction(spellObjects.CC.FreezeTrap, true);
				action.EvaluateInterruptAction(spellObjects.CC.Polymorph, true);
				action.EvaluateInterruptAction(spellObjects.CC.Supernova, true);
				action.EvaluateInterruptAction(spellObjects.CC.Paralysis, true);
				action.EvaluateInterruptAction(spellObjects.CC.Repentance, true);
				action.EvaluateInterruptAction(spellObjects.CC.ShackleUndead, true);
				action.EvaluateInterruptAction(spellObjects.CC.Blind, true);
				action.EvaluateInterruptAction(spellObjects.CC.Gouge, true);
				action.EvaluateInterruptAction(spellObjects.CC.Hex, true);
				action.EvaluateInterruptAction(spellObjects.CC.IntimidatingShout, true);

				action.EvaluateInterruptAction(spellObjects.CC.BlindingSleet, Target.InRange(12));
				action.EvaluateInterruptAction(spellObjects.CC.SigilOfMisery, Target.InRange(30));
				action.EvaluateInterruptAction(spellObjects.CC.IncapacitatingRoar, Target.InRange(10));
				action.EvaluateInterruptAction(spellObjects.CC.IntimidatingRoar, Target.InRange(10));
				action.EvaluateInterruptAction(spellObjects.CC.Typhoon, Target.InRange(15));
				action.EvaluateInterruptAction(spellObjects.CC.BlastWave, Target.InRange(8));
				action.EvaluateInterruptAction(spellObjects.CC.BlindingLight, Target.InRange(10));
				action.EvaluateInterruptAction(spellObjects.CC.PsychicScream, Target.InRange(8));
			end
		end,

		Interrupt = function(action)
			action.EvaluateInterruptAction(spellObjects.Interrupt.MindFreeze, true);
			action.EvaluateInterruptAction(spellObjects.Interrupt.Disrupt, true);
			action.EvaluateInterruptAction(spellObjects.Interrupt.SkullBash, true);
			action.EvaluateInterruptAction(spellObjects.Interrupt.SolarBeam, true);
			action.EvaluateInterruptAction(spellObjects.Interrupt.CounterShot, true);
			action.EvaluateInterruptAction(spellObjects.Interrupt.Muzzle, true);
			action.EvaluateInterruptAction(spellObjects.Interrupt.Counterspell, true);
			action.EvaluateInterruptAction(spellObjects.Interrupt.SpearHandStrike, true);
			action.EvaluateInterruptAction(spellObjects.Interrupt.Rebuke, true);
			action.EvaluateInterruptAction(spellObjects.Interrupt.Silence, true);
			action.EvaluateInterruptAction(spellObjects.Interrupt.Kick, true);
			action.EvaluateInterruptAction(spellObjects.Interrupt.WindShear, true);
			action.EvaluateInterruptAction(spellObjects.Interrupt.SpellLock, true);
			action.EvaluateInterruptAction(spellObjects.Interrupt.Pummel, true);

			action.EvaluateInterruptAction(spellObjects.Interrupt.SigilOfSilence, Target.InRange(30));
		end,

		Stun = function(action)
			local interruptManager = addonTable.interruptManager;

			if Target.IsStunnable() and coreTables.GetValue(interruptManager, "useStun", 1) == 1 then
				action.EvaluateInterruptAction(spellObjects.Stun.CheapShot, Player.IsStealthed());
				action.EvaluateInterruptAction(spellObjects.Stun.Maim, Requirements.Maim);
				action.EvaluateInterruptAction(spellObjects.Stun.BetweenTheEyes, Requirements.BetweenTheEyes);
				action.EvaluateInterruptAction(spellObjects.Stun.KidneyShot, Requirements.KidneyShot);

				action.EvaluateInterruptAction(spellObjects.Stun.Asphyxiate, true);
				action.EvaluateInterruptAction(spellObjects.Stun.Gnaw, true);
				action.EvaluateInterruptAction(spellObjects.Stun.FelEruption, true);
				action.EvaluateInterruptAction(spellObjects.Stun.MightyBash, true);
				action.EvaluateInterruptAction(spellObjects.Stun.Intimidation, true);
				action.EvaluateInterruptAction(spellObjects.Stun.QuakingPalm, true);
				action.EvaluateInterruptAction(spellObjects.Stun.HammerOfJustice, true);
				action.EvaluateInterruptAction(spellObjects.Stun.PsychicHorror, true);
				action.EvaluateInterruptAction(spellObjects.Stun.AxeToss, true);
				action.EvaluateInterruptAction(spellObjects.Stun.StormBolt, true);

				action.EvaluateInterruptAction(spellObjects.Stun.ChaosNova, Target.InRange(8));
				action.EvaluateInterruptAction(spellObjects.Stun.LegSweep, Target.InRange(5));
				action.EvaluateInterruptAction(spellObjects.Stun.CapacitorTotem, Target.InRange(8))
				action.EvaluateInterruptAction(spellObjects.Stun.Shockwave, Target.InRange(10));
				action.EvaluateInterruptAction(spellObjects.Stun.BullRush, Target.InRange(6));
				action.EvaluateInterruptAction(spellObjects.Stun.WarStomp, Target.InRange(8));
				action.EvaluateInterruptAction(spellObjects.Stun.Shadowfury, Target.InRange(35));
			end
		end,

		Valid = function()
			local interruptManager = coreTables.GetValue(addonTable.interruptManager, "InterruptOptions", {});
			local groupType = (Player.IsSolo() and "Solo") or (Player.InParty() and "Party") or (Player.InRaid() and "Raid") or "";

			return coreTables.GetValue(interruptManager, format("%s%s", "noLimits", groupType), 1) == 1
			    or (coreTables.GetValue(interruptManager, format("%s%s", "limitPlayer", groupType), 0) == 1 and Player.IsUnit("targettarget"))
			    or (coreTables.GetValue(interruptManager, format("%s%s", "limitPet", groupType), 0) == 1 and Pet.IsUnit("targettarget"))
			    or (coreTables.GetValue(interruptManager, format("%s%s", "limitFriendly", groupType), 0) == 1 and Target.IsFriendly("targettarget"));
		end,
	};

	addon.Icons.INTERRUPT = {
		Actions = function()
			return actions;
		end,

		EvaluateInterruptAction = function(Object, useObject)
			if coreObjects.IsSpell(Object) and useObject ~= nil then
				rotationIndex = rotationIndex + 1;

				if Object.IsUsable() and (useObject == true or (type(useObject) == "function" and useObject(addon.Enemies, Target))) then
					local spellID = Object.SpellID();
					local objectInfo = apiSpell.GetSpellInfo(spellID);

					if not coreTables.IsEmpty(objectInfo) then
						local iconInfo = {
							ID = spellID,
							Name = objectInfo.Name,
							Texture = objectInfo.IconID,
							Text = "",
							Type = "Spell",
							Keybind = modulesActionBar.Keybind("Spell", spellID),
							Wait = false,
							WaitStart = 0,
							WaitEnd = 0,
							WaitTime = 0;
							RotationIndex = rotationIndex,
						};

						tinsert(evaluateActions, iconInfo);
						stopped = true;
					end
				end
			end
		end,
	};

	------------------------ WOW API Events ----------------------
	Events = {
		CheckCasting = function(...)
			local _, unitID = ...;

			if not unitID or unitID == "target" then
				if Target.IsValidTarget() and Target.Casting.IsInterruptible() then
					if not systemTimers.TimerExists(Events.TimerEvent) then
						Events.TimerEvent();
						systemTimers.TimerAdd(Events.TimerEvent, addon.refreshInterval);
					end
				else
					if systemTimers.TimerExists(Events.TimerEvent) then
						systemTimers.TimerRemove(Events.TimerEvent);
						wipe(actions);

						systemCallbacks.Callback("InterruptIcon");
					end
				end
			end
		end,

		StartCasting = function(...)
			local _, unitID = ...;

			if unitID == "target" then
				if Target.IsValidTarget() and Target.Casting.IsInterruptible() then
					if not systemTimers.TimerExists(Events.TimerEvent) then
						Events.TimerEvent();
						systemTimers.TimerAdd(Events.TimerEvent, addon.refreshInterval);
					end
				else
					if systemTimers.TimerExists(Events.TimerEvent) then
						systemTimers.TimerRemove(Events.TimerEvent);
						wipe(actions);

						systemCallbacks.Callback("InterruptIcon");
					end
				end
			end
		end,

		StopCasting = function(...)
			local _, unitID = ...;

			if unitID == "target" then
				if not Target.IsValidTarget() or not Target.Casting.IsInterruptible() then
					if systemTimers.TimerExists(Events.TimerEvent) then
						systemTimers.TimerRemove(Events.TimerEvent);

						wipe(actions);

						systemCallbacks.Callback("InterruptIcon");
					end
				end
			end
		end,

		TimerEvent = function()
			if Target.IsValidTarget() and Suggest.Valid() then
				local iconInterrupt = addon.Icons.INTERRUPT;

				rotationIndex = 0;
				stopped = false;
				evaluateActions = {};

				systemCallbacks.Callback("ProcessInterrupt", iconInterrupt);
				sort(evaluateActions, sortActions);

				wipe(actions);
				actions = coreTables.Duplicate(evaluateActions);

				systemCallbacks.Callback("InterruptIcon", actions);

				evaluateActions = nil;
			else
				wipe(actions);

				systemCallbacks.Callback("InterruptIcon");
			end
		end,
	};

	----------------------- Plugin Code Operation ----------------
	local moduleEvents = {
		playerLogin = function()
			Events.CheckCasting();
		end,

		ProcessInterrupt = function(action)
			if not stopped and Target.Casting.IsInterruptible() and (coreTables.GetValue(addonTable.interruptManager, "requireRotation", 1) == 0 or addon.currentRotation ~= nil) then
				Suggest.Interrupt(action);
				Suggest.Stun(action);
				Suggest.CrowdControl(action);
			end
		end,
	};

	local tableEvents = {
		PLAYER_TARGET_CHANGED = Events.CheckCasting,
		UNIT_SPELLCAST_CHANNEL_START = Events.StartCasting,
		UNIT_SPELLCAST_CHANNEL_STOP = Events.StopCasting,
		UNIT_SPELLCAST_CHANNEL_UPDATE = Events.StartCasting,
		UNIT_SPELLCAST_FAILED = Events.StopCasting,
		UNIT_SPELLCAST_FAILED_QUIET = Events.StopCasting,
		UNIT_SPELLCAST_INTERRUPTED = Events.StopCasting,
		UNIT_SPELLCAST_START = Events.StartCasting,
		UNIT_SPELLCAST_STOP = Events.StopCasting,
		UNIT_SPELLCAST_SUCCEEDED = Events.CheckCasting,
	};

	systemCallbacks.RegisterAddonCallbacks(moduleEvents);
	systemEvents.RegisterEvents(tableEvents, true);
end