-- Pulls back the Addon-Local Variables and store them locally.
local _, addonTable = ...;
local addon = _G["LunaEclipse_Rotation"];

-- Plugin information variables.
local pluginName = "Macro Manager";
local pluginDescription = "This plugin changes the players character specific macros based on specialization.\n\nThis allows a total of 18 character specific macros per specialization!";

-- Check for LunaEclipse Rotation Helper being loaded first so it doesn't register the plugin if its not needed.
if addon then
    local optionsTable = {
        type = "group",
        name = pluginName,
        get = function(info)
            local returnValue = 0;

            if info then
                local Tables = addon.Core.Tables;
                local profile = addonTable.addonSettings;
                local keyName = info[#info];

                returnValue = Tables.GetValue(profile, keyName);

                if returnValue == nil then
                    returnValue = 0;
                    profile[keyName] = returnValue;

                    addonTable.addonSettings = profile;
                end
            end

            return addon.Options.Core.ConvertToUI(returnValue, info.option.tristate);
        end,
        set = function(info, value)
            if info then
                local profile = addonTable.addonSettings;
                local keyName = info[#info];

                if info.option.type == "toggle" then
                    value = addon.Options.Core.ConvertToValue(value, info.option.tristate);
                end

                profile[keyName] = value;

                -- Put the settings into the database
                addonTable.addonSettings = profile;
            end
        end,
        childGroups = "tree",
        args = {
            allowZeroMacros = {
                type = "toggle",
                name = "Allow Macro Profiles with Zero Macros",
                desc = "This allows a profile to be considered valid when it contains no macros, unchecking this will cause the addon to treat profiles with no macros as not existing.",
                width = "full",
                order = 10,
            },
        },
    };

    -- The plugin still has to be registered even if its not loaded into memory, otherwise there is no checkbox to re-enable it.
    addon.System.Plugins.RegisterPlugin(pluginName, pluginDescription, optionsTable);
end
