-- Pulls back the Addon-Local Variables and store them locally.
local addonName, addonTable = ...;
local addon = _G["LunaEclipse_Rotation"];

-- Plugin information variables.
local pluginName = "Talent Builds";

-- Check whether the plugin is supposed to be active
if addon and addon.System.Plugins.IsActive(pluginName) then
	-- Store local copies of Blizzard API and other addon global functions and variables.
	local gsub, pairs, select, string, tonumber, tostring, unpack = gsub, pairs, select, string, tonumber, tostring, unpack;

	-- Tome and Codex buffs to check and the max valid level.
	local TOME_BUFF_AURA = {
		[227564] = 100,										-- Codex of the Clear Mind (Level 100)
		[226234] = addonTable.TalentBuilds.MaxLevel,		-- Codex of the Tranquil Mind (Max Level)
		[227563] = 109,										-- Tome of the Clear Mind (Level 109)
		[227041] = addonTable.TalentBuilds.MaxLevel,		-- Tome of the Tranquil Mind (Max Level)
	};

	-- Preparation buffs to check.
	local PREPARATION_AURA = {
		[32727] = true,			-- Preparation Buff (Arena)
		[228128] = true,		-- Preparation Buff (Dungeon)
	};

	-- Table to store changing talents queue, and the queued builds name.
	local queuedBuild, talentQueue;

	-- ElvUI variables only...
	local Skins, useElvUI;

	-- Variables to store information for UI elements.
	local buttonHeight, buttonSeperator, buttonWidth = 23, 10, 80;

	-- Tables to hold private functions.
	local Buffs, Builds, Conversion, Events, Hooks, Queue, UICreate, UIEvents = {}, {}, {}, {}, {}, {}, {}, {};

	--------------------- Popup Dialog Boxes ---------------------
	StaticPopupDialogs["LERH_ADD_BUILD"] = {
		text = "Enter Build Name:",
		button1 = "Save",
		button2 = "Cancel",
		OnAccept = function(self)
			local Tables = addon.Core.Tables;
			local TalentBuilds = addon.Features.TalentBuilds;

			-- Get the current specialization group.
			local specInfo = addon.API.General.GetSpecializationInfo();
			local specGroup = addon.API.Blizzard.GetActiveSpecGroup();

			-- Check to see that the player has an active specialization.
			if specGroup then
				local _, presetBuilds, customBuilds = Builds.AvailableBuilds();
				local buildName = self.editBox:GetText();

				-- Get the players current talents as a table.
				local currentTalents = TalentBuilds.CurrentTalents();
				-- Convert the talent table into a talent string.
				local talentString = Conversion.FromTalents(currentTalents, specGroup);

				if Tables.GetValue(presetBuilds, buildName) or Tables.GetValue(customBuilds, buildName) then
					-- Build exists, prompt to overwrite/override
					local dialog = addon.API.Blizzard.StaticPopup_Show("LERH_OVERWRITE_BUILD", buildName);

					if dialog then
						dialog.data = buildName;
						dialog.data2 = talentString;
					end
				else
					customBuilds[buildName] = talentString;

					addonTable.talentBuilds.CustomBuilds[specInfo.Name] = customBuilds;
				end

				addonTable.TalentBuilds.createBuildUI();
			end
		end,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
		preferredIndex = 3,
		hasEditBox = true,
	};

	StaticPopupDialogs["LERH_OVERWRITE_BUILD"] = {
		text = "%s build already exists, do you wish to overwrite it?",
		button1 = "Yes",
		button2 = "No",
		OnAccept = function(self, data, data2)
			local specInfo = addon.API.General.GetSpecializationInfo();
			local _, _, customBuilds = Builds.AvailableBuilds();

			customBuilds[data] = data2;

			addonTable.talentBuilds.CustomBuilds[specInfo.Name] = customBuilds;
			addonTable.TalentBuilds.createBuildUI();
		end,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
		preferredIndex = 3,
	};

	StaticPopupDialogs["LERH_REMOVE_BUILD"] = {
		text = "Do you want to remove the build %s?",
		button1 = "Yes",
		button2 = "No",
		OnAccept = function()
			local specInfo = addon.API.General.GetSpecializationInfo();
			local buildName = addon.API.Blizzard.UIDropDownMenu_GetSelectedName(LERH_BuildList);
			local _, _, customBuilds = Builds.AvailableBuilds();

			customBuilds[buildName] = nil;

			addonTable.talentBuilds.CustomBuilds[specInfo.Name] = customBuilds;
			addonTable.TalentBuilds.createBuildUI();
		end,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
		preferredIndex = 3,
	};

	StaticPopupDialogs["LERH_REMOVE_BUILD_ERROR"] = {
		text = "The %s build is a protected preset build and can't be deleted.",
		button1 = "OK",
		timeout = 15,
		whileDead = true,
		hideOnEscape = true,
	};

	StaticPopupDialogs["LERH_TALENT_CHANGES_FAILED"] = {
		text = string.join("", "Unable to apply the %s build as the following talents, %s were locked.", "\n\n", "This is most likely because a ", addonTable.COLORS.TAGS.ACTIVE, "related spell is on cooldown", addonTable.TAGS.END, ".", "\n\n", "Please try again later!"),
		button1 = "OK",
		timeout = 15,
		whileDead = true,
		hideOnEscape = true,
	};

	StaticPopupDialogs["LERH_TALENT_CHANGES_NO_TALENT_BUILD"] = {
		text = "No talent build specified!",
		button1 = "OK",
		timeout = 15,
		whileDead = true,
		hideOnEscape = true,
	};

	StaticPopupDialogs["LERH_TALENT_CHANGES_NOT_PERMITTED"] = {
		text = "Talent changes are not currently permitted!",
		button1 = "OK",
		timeout = 15,
		whileDead = true,
		hideOnEscape = true,
	};

	StaticPopupDialogs["LERH_TALENT_CHANGES_TALENT_BUILD_UNAVAILABLE"] = {
		text = "Talent build not available!",
		button1 = "OK",
		timeout = 15,
		whileDead = true,
		hideOnEscape = true,
	};

	StaticPopupDialogs["LERH_UPDATE_BUILD"] = {
		text = "You have new unlearned talents, do you wish to learn the appropriate talents for the %s build?",
		button1 = "Yes",
		button2 = "No",
		OnAccept = function(self, data, data2)
			Builds.Update(data2);
		end,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
		preferredIndex = 3,
	};

	--------------------- Private Module Functions ---------------
	Buffs = {
		PreperationActive = function()
			local apiUnit = addon.API.Unit;

			-- Loop through the array of preparation buffs to check.
			for spellID, enabled in pairs(PREPARATION_AURA) do
				-- If a valid buff is found return true, we don't need to check any more
				if apiUnit.UnitAura("player", spellID, true) then
					return true;
				end
			end

			-- Return false because the buff wasn't found.
			return false;
		end,

		TomeActive = function()
			local apiUnit = addon.API.Unit;
			local Tables = addon.Core.Tables;

			-- Get the players level to see if the buff is valid on them.
			local unitInfo = apiUnit.GetUnitInfo("player");

			-- Loop through the array of tome and codex buffs to check.
			for spellID, requiredLevel in pairs(TOME_BUFF_AURA) do
				-- If a valid buff is found return true, we don't need to check any more
				if apiUnit.UnitAura("player", spellID, true) and Tables.GetValue(unitInfo, "UnitLevel", addonTable.TalentBuilds.MaxLevel) <= requiredLevel then
					return true;
				end
			end

			-- Return false because the buff wasn't found.
			return false;
		end,
	};

	Builds = {
		Apply = function(buildTalents)
			-- Check that talents information was given
			if buildTalents then
				-- Get the current specialization group.
				local specGroup = addon.API.Blizzard.GetActiveSpecGroup();

				-- Get the maximum row the player can select talents from.
				local numRows = addon.API.Blizzard.GetMaxTalentTier();
				local selected, tier, column;

				-- Loop through all talents in the table.
				for talentID, talentName in pairs(buildTalents) do
					-- Get whether the talent is already selected.
					selected = select(4, addon.API.Blizzard.GetTalentInfoByID(talentID, specGroup));
					-- Get the rown and column of the talent position.
					tier, column = select(8, addon.API.Blizzard.GetTalentInfoByID(talentID, specGroup));

					-- Check to make sure the talent doesn't require a higher level to learn.
					if tier <= numRows then
						-- Check to see that the talent is not already selected.
						if not selected then
							-- It's not selected and we have the required level so add it to the queue.
							Queue.AddTalent(talentID, talentName);
						end
					end
				end
			end
		end,

		AvailableBuilds = function()
			local Tables = addon.Core.Tables;

			-- Get the current specialization info
			local classInfo = addon.API.Unit.UnitClass();
			local specInfo = addon.API.General.GetSpecializationInfo();

			-- Get any saved builds, this is custom builds or altered preset builds
			local customBuilds = addonTable.talentBuilds.CustomBuilds[specInfo.Name] or {};

			-- Get any builds from the rotation file.
			local rotationBuilds = addon.currentRotation ~= nil and addon.currentRotation.presetBuilds or {};

			-- Get any preset builds
			local classBuilds = addonTable.TalentBuilds.Builds[Tables.GetValue(classInfo, "ID", 0)] or {}
			local presetBuilds = classBuilds[Tables.GetValue(specInfo, "ID", 0)] or {};

			presetBuilds = Tables.Merge(presetBuilds, rotationBuilds);

			-- Combine the two tables to get all builds, Custom Builds overwrite rotation builds when they have the same name.
			local combinedBuilds = Tables.Merge(presetBuilds, customBuilds);

			-- Return the builds available for the players specialization.
			return combinedBuilds, presetBuilds, customBuilds;
		end,

		BuildName = function(talentString, specGroup, requireComplete)
			local Tables = addon.Core.Tables;

			-- Check to make sure we have been given talents and specialization
			if talentString and specGroup then
				-- Gets the available builds for the current specialization.
				local _, presetBuilds, customBuilds = Builds.AvailableBuilds();

				-- Check to see if we actually found some custom builds.
				if not Tables.IsEmpty(customBuilds) then
					-- Loop through every build in the table.
					for buildName, buildTalentString in pairs(customBuilds) do
						-- Check to see if the stored talent string matches the talent string for the current talents.
						if Conversion.Equate(talentString, buildTalentString, requireComplete) then
							-- It matches so set the return value to the build name.
							return buildName;
						end
					end
				end

				-- Check to see if we actually found some preset builds.
				if not Tables.IsEmpty(presetBuilds) then
					-- Loop through every build in the table.
					for buildName, buildTalentString in pairs(presetBuilds) do
						-- Check to see if the stored talent string matches the talent string for the current talents.
						if Conversion.Equate(talentString, buildTalentString) then
							-- It matches so set the return value to the build name.
							return buildName;
						end
					end
				end
			end

			-- Return the build name if found, otherwise return "Custom Build" if there was no match.
			return addonTable.TalentBuilds.Custom;
		end,

		BuildNameFromTalents = function()
			local TalentBuilds = addon.Features.TalentBuilds;

			-- Get the current specialization group.
			local specGroup = addon.API.Blizzard.GetActiveSpecGroup();

			-- Check to see that the player has an active specialization.
			if specGroup then
				-- Get the players current talents as a table.
				local currentTalents = TalentBuilds.CurrentTalents();
				-- Convert the talent table into a talent string.
				local talentString = Conversion.FromTalents(currentTalents, specGroup);

				-- Return the build name based on the talent string.
				return Builds.BuildName(talentString, specGroup);
			end

			-- Return the build name if found, otherwise return "Custom Build" if there was no match.
			return addonTable.TalentBuilds.Custom;
		end,

		-- Used to select a build from the build list
		Select = function(buildName)
			if LERH_BuildList then
				-- Get the current specialization info.
				local specInfo = addon.API.General.GetSpecializationInfo();

				if not buildName then
					local TalentBuilds = addon.Features.TalentBuilds;
					local buildInfo = TalentBuilds.SelectedBuild();

					buildName = buildInfo.Name;
				end

				addonTable.talentBuilds.CurrentBuild[specInfo.Name] = buildName;

				addon.API.Blizzard.UIDropDownMenu_SetSelectedName(LERH_BuildList, buildName);
				addon.API.Blizzard.UIDropDownMenu_SetText(LERH_BuildList, buildName);

				-- Send a callback that the build has been updated so minimap icon can change information.
				addon.System.Callbacks.Callback("updateMinimapIcon");
			end
		end,

		Update = function(talentString)
			-- Check that talents information was given
			if talentString then
				-- Get the current specialization group.
				local specGroup = addon.API.Blizzard.GetActiveSpecGroup();

				-- Get the maximum row the player can select talents from.
				local numRows = addon.API.Blizzard.GetMaxTalentTier();
				local available, selected, column, talentID, talentName;

				-- We are about to start learning talents so points will change, lets stop listening to the event until the changes are finished.
				addon.System.Events.EventRemove("PLAYER_CHARACTER_UPGRADE_TALENT_COUNT_CHANGED", Events.availablePointsChanged, true);

				for row = 1, numRows do
					-- Set the column based on the value in the talent string.
					column = tonumber(string.sub(talentString, row, row));

					-- Make sure a valid column is specified.
					if column >= 1 and column <= 3 then
						available, selected = addon.API.Blizzard.GetTalentTierInfo(row, specGroup);

						if available and selected == 0 then
							-- Get the infomation about the talent being checked.
							talentID, talentName = addon.API.Blizzard.GetTalentInfo(row, column, specGroup);

							-- It's not selected and we have the required level so add it to the queue.
							Queue.AddTalent(talentID, talentName);
						end
					end
				end
			end
		end,
	};

	Conversion = {
		-- Check to see if two talent strings match based on Optional talents, and number of talent tiers
		-- available and will specify a match if player has unspent talent points as long as the selected
		-- talents match unless requireComplete is specified.
		Equate = function(currentTalentString, buildTalentString, requireComplete)
			-- Set the return value to false by default.
			local returnValue = false;
			local counter = 0;

			-- Check to see if we were supplied both a build talent string and preset talent string.
			if currentTalentString and buildTalentString then
				-- We have two strings to compare, so lets start by presuming they match.
				returnValue = true;

				-- Get the max number of talent rows to check.
				local numRows = addon.API.Blizzard.GetMaxTalentTier();
				-- Variables to hold the numbers to compare.
				local currentTalent, buildTalent;

				-- Loop through all the rows the player can choose talents from.
				for row = 1, numRows do
					-- Get the column number of the talent from the current build.
					currentTalent = tonumber(string.sub(currentTalentString, row, row));
					-- Get the column number of the talent from the preset build.
					buildTalent = tonumber(string.sub(buildTalentString, row, row));
					-- Increase the selected talent counter if needed
					counter = currentTalent ~= 0 and counter + 1 or counter;

					-- Check to see if the talents don't match, and a talent is specified in the preset build.
					if ((requireComplete == nil and currentTalent ~= 0) or requireComplete ~= nil) and currentTalent ~= buildTalent and buildTalent ~= 0 then
						-- They don't match so set the returnValue to false and exit the loop as there is no need to check any more.
						returnValue = false;
						break;
					end
				end

				-- If there is no active talents then the string can't be considered a match as it would match everything.
				if counter == 0 then
					returnValue = false;
				end
			end

			-- Return whether the strings match, will also return false if no strings are provided.
			return returnValue;
		end,

		-- Create a talent string from a table of talents
		FromTalents = function(buildTalents, specGroup)
			-- Set default return value with no talents chosen.
			local returnValue = "0000000";

			if buildTalents and specGroup then
				local tier, column;

				-- Loop through all talents in the table.
				for talentID in pairs(buildTalents) do
					-- Get the rown and column of the talent position.
					tier, column = select(8, addon.API.Blizzard.GetTalentInfoByID(talentID, specGroup));

					-- Modify the return value with the new information.
					returnValue = string.format("%s%s%s", string.sub(returnValue, 1, tier - 1), tostring(column), string.sub(returnValue, tier + 1));
				end
			end

			-- Return the completed string.
			return returnValue;
		end,

		-- Create a table of talents from a talent string
		ToTalents = function(talentString, specGroup)
			local returnValue;

			-- Check to see if a talent string and specialization was given.
			if talentString and specGroup then
				local talentID, name;
				-- Get the number of rows in the talent string.
				local numRows = string.len(talentString);
				local column;

				-- Loop through the rows in the talent string.
				for row = 1, numRows do
					-- Set the column based on the value in the talent string.
					column = tonumber(string.sub(talentString, row, row));

					-- Make sure a valid column is specified.
					if column >= 1 and column <= 3 then
						-- Get the infomation about the talent being checked.
						talentID, name = addon.API.Blizzard.GetTalentInfo(row, column, specGroup);

						-- Check to see if the returnValue contains data.
						if not returnValue then
							-- It doesn't contain data so set it up as a new table.
							returnValue = {};
						end

						-- Store the talent name using the ID as the table key.
						returnValue[talentID] = name;
					end
				end
			end

			-- Return the talent table or nil if the player has no specialization.
			return returnValue;
		end,
	};

	Queue = {
		AddTalent = function(talentID, talentName)
			talentQueue = talentQueue or {};

			-- Check to see if we were passed a talent.
			if talentID and talentName then
				-- Add it to the queue with a value of false meaning it hasn't been checked.
				talentQueue[talentID] = { Name = talentName };
				-- Try to learn the talent.
				addon.API.Blizzard.LearnTalent(talentID);
			end
		end,

		Count = function()
			-- Set the default return value.
			local returnValue = 0;

			-- Loop through all talents in the changes queue.
			for _, data in pairs(talentQueue) do
				-- Check to see if the talent failed.
				if not data.failed then
					-- Only increase the counter if did not fail.
					returnValue = returnValue + 1;
				end
			end

			-- Return the count.
			return returnValue;
		end,

		Failed = function(talentID)
			-- Check to see if we were passed a talent, and that it is a talent currently in the queue.
			if talentID and talentQueue[talentID] then
				-- Mark it as failed because it was a talent we were tracking.
				talentQueue[talentID].failed = true;
			end
		end,

		FailedCount = function()
			-- Set the default return value.
			local returnValue = 0;

			-- Loop through all talents in the changes queue.
			for _, data in pairs(talentQueue) do
				-- Check to see if the talent failed.
				if data.failed then
					-- Only increase the counter if failed.
					returnValue = returnValue + 1;
				end
			end

			-- Return the count.
			return returnValue;
		end,

		FailedDisplay = function()
			-- Get the count of the total number of talent changes that failed.
			local failedTalents = Queue.FailedCount();

			local talentString;
			local counter = 0;

			-- Loop through all the talents in the queue, these should all be failed as successful ones are removed.
			for _, data in pairs(talentQueue) do
				-- Increase the counter.
				counter = counter + 1;

				if counter == 1 then
					-- First entry so just store the name.
					talentString = string.format("%s%s%s", addonTable.COLORS.TAGS.ACTIVE, data.Name, addonTable.TAGS.END);
				elseif counter == failedTalents then
					-- Last entry and there is more then one, so use "and" to join the new name.
					talentString = string.format("%s and %s%s%s", talentString, addonTable.COLORS.TAGS.ACTIVE, data.Name, addonTable.TAGS.END);
				else
					-- Not first or last entry so seperate with commas.
					talentString = string.format("%s, %s%s%s", talentString, addonTable.COLORS.TAGS.ACTIVE, data.Name, addonTable.TAGS.END);
				end
			end

			-- Display the popup window with the talent names that failed.
			addon.API.Blizzard.StaticPopup_Show("LERH_TALENT_CHANGES_FAILED", string.format("%s%s%s", addonTable.COLORS.TAGS.ACTIVE, queuedBuild, addonTable.TAGS.END), talentString);

			-- Delete the talent queue now that the popup message box has been displayed.
			talentQueue = nil;
		end,

		RemoveTalent = function(talentID)
			-- Check to see if we were passed a talent, and that it is a talent currently in the queue.
			if talentID and talentQueue[talentID] then
				-- Set the entry to nil to remove it from the queue.
				talentQueue[talentID] = nil;
			end
		end,

		Verify = function()
			local uncheckedTalents = Queue.Count();
			local failedTalents = Queue.FailedCount();

			-- Check to see that all talents in the queue have been checked and at least one talent failed.
			if queuedBuild and uncheckedTalents == 0 then
				if failedTalents >= 1 then
					-- Display the popup window displaying the talents that failed.
					Queue.FailedDisplay();
				end

				local buildName = Builds.BuildNameFromTalents();

				Builds.Select(buildName);

				-- Reset queued builds name, as we have now finished the queue.
				queuedBuild = nil;

				-- Finished the add queue, now re-enable check for change in talent points.
				addon.System.Events.EventAdd("PLAYER_CHARACTER_UPGRADE_TALENT_COUNT_CHANGED", Events.availablePointsChanged, true);
			end
		end,
	};

	------------------- WOW API Events and Hooks -----------------
	Events = {
		CheckAllowed = function(...)
			local TalentBuilds = addon.Features.TalentBuilds;
			local LERH_Build, LERH_BuildApply, LERH_BuildList, LERH_BuildSave, LERH_BuildRemove = LERH_Build, LERH_BuildApply, LERH_BuildList, LERH_BuildSave, LERH_BuildRemove;
			local event, unitID = ...;

			if LERH_Build and (event ~= "UNIT_AURA" or string.lower(unitID) == "player") then
				if TalentBuilds.ChangesAllowed() then
					addon.API.Blizzard.UIDropDownMenu_EnableDropDown(LERH_BuildList);

					LERH_BuildApply:Enable();
					LERH_BuildSave:Enable();
					LERH_BuildRemove:Enable();
				else
					addon.API.Blizzard.UIDropDownMenu_DisableDropDown(LERH_BuildList);

					LERH_BuildApply:Disable();
					LERH_BuildSave:Disable();
					LERH_BuildRemove:Disable();
				end
			end
		end,

		PointsChanged = function(...)
			-- Check to see if there is new points available
			if addon.API.Blizzard.GetNumUnspentTalents() > 0 then
				addon.Features.TalentBuilds.Update();
			end
		end,

		TalentFailed = function()
			-- Check to see there is even a queue to check
			if talentQueue then
				-- Get the list of talents that failed since the last clear.
				local failedTalents = addon.API.Blizzard.GetFailedTalentIDs();
				local talentID;

				-- Loop through the failed talents.
				for counter = 1, #failedTalents do
					-- Get the talent ID.
					talentID = failedTalents[counter];

					-- Mark the talent as failed, this only happens if it is already in the queue.
					Queue.Failed(talentID);
				end

				-- Clear the list of failed talents.
				addon.API.Blizzard.ClearFailedTalentIDs();

				-- Check if the talent queue has been completed.
				Queue.Verify();
			end
		end,

		TalentUpdate = function()
			-- Check to see there is even a queue to check
			if talentQueue then
				-- Get the current specialization group.
				local specGroup = addon.API.Blizzard.GetActiveSpecGroup();

				-- We have no way of telling which talent succeeded, so loop through all the talents in the queue.
				for talentID, data in pairs(talentQueue) do
					-- Check to see if the talent has not already been marked as failed.
					if not data.failed then
						-- Check to see if the talent in the queue is now selected.
						if select(4, addon.API.Blizzard.GetTalentInfoByID(talentID, specGroup)) then
							-- The talent is now selected so remove it from the queue.
							Queue.RemoveTalent(talentID);
						end
					end
				end

				-- Check if the talent queue has been completed.
				Queue.Verify();
			else
				local buildName = Builds.BuildNameFromTalents();

				Builds.Select(buildName);
			end
		end,
	};

	Hooks = {
		-- Fired when the talent frame changes tabs
		ChangeTab = function(...)
			-- This is raw hooked, so call the original function
			addon.hooks.PanelTemplates_SetTab(...);

			-- Don't continue if the player doesn't have a talent frame yet (under level 10)
			if PlayerTalentFrame then
				-- Only show when the talents tab is open
				if addon.API.Blizzard.PanelTemplates_GetSelectedTab(PlayerTalentFrame) == 2 then
					-- Build the frame
					UICreate.UserInterface();

					-- Set the visibility of the profile selector to that of the talent frame
					LERH_Build:Show();
				else
					if LERH_Build then
						LERH_Build:Hide();
					end
				end
			end
		end,

		-- Fired when talent the talent frame is toggled
		ToggleFrame = function()
			-- This is raw hooked, so call the original function
			addon.hooks.ToggleTalentFrame();

			-- Don't continue if the player doesn't have a talent frame yet (under level 10)
			if PlayerTalentFrame then
				if addon.API.Blizzard.PanelTemplates_GetSelectedTab(PlayerTalentFrame) == 2 then -- Only show when the talents tab is open
					-- Build the frame
					UICreate.UserInterface();

					-- Set the visibility of the profile selector to that of the talent frame
					LERH_Build:Show();
				else
					if LERH_Build then
						LERH_Build:Hide();
					end
				end
			end
		end,
	};

	---------------------- UI Creation and Events ----------------
	UICreate = {
		Apply = function()
			local TalentBuilds = addon.Features.TalentBuilds;

			if not LERH_BuildApply then
				local applyButton = addon.API.Blizzard.CreateFrame("Button", "LERH_BuildApply", LERH_Build, "UIPanelButtonTemplate");
				applyButton:SetSize(buttonWidth, buttonHeight);
				applyButton:SetText("Apply");

				if useElvUI then
					applyButton:SetPoint("TOPLEFT", LERH_BuildList, "TOPRIGHT", buttonSeperator, 0);

					Skins:HandleButton(applyButton);
				else
					applyButton:SetPoint("TOPLEFT", LERH_BuildList, "TOPRIGHT", buttonSeperator, 0);
				end

				if TalentBuilds.ChangesAllowed() then
					applyButton:Enable();
				else
					applyButton:Disable();
				end

				applyButton:SetScript("OnClick", UIEvents.Activate);
				applyButton:Show();
			end
		end,

		Builds = function()
			local TalentBuilds = addon.Features.TalentBuilds;

			if not LERH_BuildList then
				local buildList = addon.API.Blizzard.CreateFrame("Button", "LERH_BuildList", LERH_Build, "UIDropDownMenuTemplate");

				if useElvUI then
					buildList:SetPoint("TOPLEFT", LERH_Build, "TOPLEFT", 100, -13);
					-- ElvUI dropdown skinning is bugged and makes the arrow button far too wide
					Skins:HandleDropDownBox(buildList);
					-- so make it the correct size
					LERH_BuildListButton:SetWidth(LERH_BuildListButton:GetHeight());

					-- and allow the user to open the dropdown by clicking anywhere on the control
					buildList:SetScript("OnClick", function(...) LERH_BuildListButton:Click() end);
					-- make the dropdown the same height as the buttons
					buildList:SetHeight(buttonHeight);
				else
					addon.API.Blizzard.UIDropDownMenu_SetWidth(buildList, 125);
					addon.API.Blizzard.UIDropDownMenu_SetButtonWidth(buildList, 125);

					buildList:SetPoint("TOPLEFT", LERH_Build, "TOPLEFT", 0, -13);
				end

				if TalentBuilds.ChangesAllowed() then
					addon.API.Blizzard.UIDropDownMenu_EnableDropDown(buildList);
				else
					addon.API.Blizzard.UIDropDownMenu_DisableDropDown(buildList);
				end

				buildList:Show();
			end

			if LERH_BuildList then
				-- Repopulate the dropdown, even if it already exists
				addon.API.Blizzard.UIDropDownMenu_Initialize(LERH_BuildList, UICreate.BuildsInitialize);
				addon.API.Blizzard.UIDropDownMenu_JustifyText(LERH_BuildList, "LEFT");

				Builds.Select();
			end
		end,

		BuildsInitialize = function(self, level)
			local allBuilds = Builds.AvailableBuilds();

			for buildName, buildString in pairs(allBuilds) do
				local info = addon.API.Blizzard.UIDropDownMenu_CreateInfo();
				info.text = buildName;
				info.value = buildString;
				info.func = UIEvents.ExistingBuild;

				addon.API.Blizzard.UIDropDownMenu_AddButton(info, level);
			end

			-- Add the new build item
			local info = addon.API.Blizzard.UIDropDownMenu_CreateInfo();
			info.text = addonTable.TalentBuilds.Custom;
			info.value = "new";
			info.func = UIEvents.CustomBuild;
			info.rgb = { 1.0, 0.0, 1.0, 1.0 };

			addon.API.Blizzard.UIDropDownMenu_AddButton(info, level);
		end,

		Frame = function()
			if not LERH_Build then
				local mainFrame = addon.API.Blizzard.CreateFrame("Frame", "LERH_Build", PlayerTalentFrameTalents);
				mainFrame.texture = mainFrame:CreateTexture(nil, "ARTWORK");
				mainFrame.texture:SetPoint("TOPLEFT");

				if useElvUI then
					mainFrame:SetSize(650, 50); -- Add 15, 10
					mainFrame:SetPoint("TOPLEFT", PlayerTalentFrame, "TOPLEFT", 5, -20);
					mainFrame.texture:SetTexture(string.format("Interface\\AddOns\\%s\\Textures\\Talent_ElvUI", addonName));
				else
					mainFrame:SetSize(600, 60); -- Add 15, 10
					mainFrame:SetPoint("TOPLEFT", PlayerTalentFrame, "TOPLEFT", 52, -20);
					mainFrame.texture:SetTexture(string.format("Interface\\AddOns\\%s\\Textures\\Talent_Standard", addonName));
				end

				mainFrame:Show();
			end
		end,

		Remove = function()
			local TalentBuilds = addon.Features.TalentBuilds;

			if not LERH_BuildRemove then
				local removeButton = addon.API.Blizzard.CreateFrame("Button", "LERH_BuildRemove", LERH_Build, "UIPanelButtonTemplate");
				removeButton:SetSize(buttonWidth, buttonHeight);
				removeButton:SetText("Remove");

				if useElvUI then
					removeButton:SetPoint("TOPLEFT", LERH_BuildSave, "TOPRIGHT", buttonSeperator, 0);

					Skins:HandleButton(removeButton);
				else
					removeButton:SetPoint("TOPLEFT", LERH_BuildSave, "TOPRIGHT", buttonSeperator, 0);
				end

				if TalentBuilds.ChangesAllowed() then
					removeButton:Enable();
				else
					removeButton:Disable();
				end

				removeButton:SetScript("OnClick", UIEvents.Remove);
				removeButton:Show();
			end
		end,

		Save = function()
			local TalentBuilds = addon.Features.TalentBuilds;

			if not LERH_BuildSave then
				local saveButton = addon.API.Blizzard.CreateFrame("Button", "LERH_BuildSave", LERH_Build, "UIPanelButtonTemplate");
				saveButton:SetSize(buttonWidth, buttonHeight); -- was w100
				saveButton:SetText("Save");

				if useElvUI then
					saveButton:SetPoint("TOPLEFT", LERH_BuildApply, "TOPRIGHT", buttonSeperator, 0);

					Skins:HandleButton(saveButton);
				else
					saveButton:SetPoint("TOPLEFT", LERH_BuildApply, "TOPRIGHT", buttonSeperator, 0);
				end

				if TalentBuilds.ChangesAllowed() then
					saveButton:Enable();
				else
					saveButton:Disable();
				end

				saveButton:SetScript("OnClick", UIEvents.Save);
				saveButton:Show();
			end
		end,

		UserInterface = function()
			UICreate.Frame();

			-- Set up the build list drop down
			UICreate.Builds()

			-- Set up the action buttons
			UICreate.Apply();
			UICreate.Save();
			UICreate.Remove();
		end,
	};

	UIEvents = {
		-- Fired when the "Activate" button is clicked
		Activate = function()
			local Tables = addon.Core.Tables;
			local TalentBuilds = addon.Features.TalentBuilds;

			local allBuilds = Builds.AvailableBuilds();
			local buildName = addon.API.Blizzard.UIDropDownMenu_GetSelectedName(LERH_BuildList);

			if Tables.GetValue(allBuilds, buildName) and buildName ~= addonTable.TalentBuilds.Custom then
				TalentBuilds.Apply(buildName);
			end
		end,

		-- Fired when Custom Build drop down item is clicked
		CustomBuild = function()
			addon.API.Blizzard.StaticPopup_Show("LERH_ADD_BUILD");
		end,

		-- Fired when an existing build drop down item is clicked
		ExistingBuild = function(self)
			Builds.Select(self:GetText());
		end,

		-- Fired when the "Remove" button is clicked
		Remove = function()
			local Tables = addon.Core.Tables;

			local _, _, customBuilds = Builds.AvailableBuilds();
			local buildName = addon.API.Blizzard.UIDropDownMenu_GetSelectedName(LERH_BuildList);

			-- Check if build is a custom build as preset builds can't be deleted
			if not Tables.IsEmpty(customBuilds) and Tables.GetValue(customBuilds, buildName) then
				addon.API.Blizzard.StaticPopup_Show("LERH_REMOVE_BUILD", buildName);
			else
				-- Build doesn't exist or is a protected preset build
				addon.API.Blizzard.StaticPopup_Show("LERH_REMOVE_BUILD_ERROR", buildName);
			end
		end,

		-- Fired when the "Save" button is clicked
		Save = function()
			local Tables = addon.Core.Tables;
			local TalentBuilds = addon.Features.TalentBuilds;

			-- Get the current specialization info and group.
			local specInfo = addon.API.General.GetSpecializationInfo();
			local specGroup = addon.API.Blizzard.GetActiveSpecGroup();

			-- Check to see that the player has an active specialization.
			if specGroup then
				local _, presetBuilds, customBuilds = Builds.AvailableBuilds();
				local buildName = addon.API.Blizzard.UIDropDownMenu_GetSelectedName(LERH_BuildList);

				-- Get the players current talents as a table.
				local currentTalents = TalentBuilds.CurrentTalents();
				-- Convert the talent table into a talent string.
				local talentString = Conversion.FromTalents(currentTalents, specGroup);

				if buildName ~= addonTable.TalentBuilds.Custom and (Tables.GetValue(presetBuilds, buildName) or Tables.GetValue(customBuilds, buildName)) then
					customBuilds[buildName] = talentString;

					addonTable.talentBuilds.CustomBuilds[specInfo.Name] = customBuilds;
				end
			end
		end,
	};

	---------------------- Public Functions ----------------------
	addon.Features.TalentBuilds = {
		ActiveBuild = function()
			local self = addon.Features.TalentBuilds;

			-- Get the current specialization group.
			local specGroup = addon.API.Blizzard.GetActiveSpecGroup();
			local returnValue;

			-- Check to see that the player has an active specialization.
			if specGroup then
				-- Get the players current talents as a table.
				local currentTalents = self.CurrentTalents();
				-- Convert the talent table into a talent string.
				local talentString = Conversion.FromTalents(currentTalents, specGroup);
				-- Get the saved build name or the name of a build that matches current talents if a default build has not been selected.
				local buildName = Builds.BuildName(talentString, specGroup);

				-- Set the return value data for the build.
				returnValue = {
					Name = buildName,
					Talents = currentTalents,
				};
			end

			-- Return the infomation found, or nil if the player does not have a specialization.
			return returnValue;
		end,

		Apply = function(buildName)
			local self = addon.Features.TalentBuilds;

			-- Store whether talent changes are currently allowed.
			local changesAllowed = self.ChangesAllowed();

			if changesAllowed and buildName and buildName ~= "" then
				-- Get a talent table of the talents associated with the build name.
				local buildTalents = self.BuildTalents(buildName);

				if buildTalents then
					-- Strip any quotes from the build name just in case.
					queuedBuild = gsub(buildName, "\"", "");

					-- Learn the talents for the build name specified.
					Builds.Apply(buildTalents);
				else
					-- There is no talent information for the build name so display the popup message box.
					addon.API.Blizzard.StaticPopup_Show("LERH_TALENT_CHANGES_TALENT_BUILD_UNAVAILABLE");
				end
			elseif not changesAllowed then
				-- Changes are not currently allowed, so display the popup message box.
				addon.API.Blizzard.StaticPopup_Show("LERH_TALENT_CHANGES_NOT_PERMITTED");
			else
				-- No build name was given so display the popup message box.
				addon.API.Blizzard.StaticPopup_Show("LERH_TALENT_CHANGES_NO_TALENT_BUILD");
			end
		end,

		Available = function()
			-- Return the builds available for the players specialization, will return nil if there are no builds.
			return Builds.AvailableBuilds();
		end,

		BuildTalents = function(buildName)
			local returnValue;

			-- Check to see if we have a build name and make sure its not empty.
			if buildName and buildName ~= "" then
				local self = addon.Features.TalentBuilds;

				-- Get the current specialization group.
				local specGroup = addon.API.Blizzard.GetActiveSpecGroup();

				-- Strip any quotes from the build name just in case.
				buildName = gsub(buildName, "\"", "");

				-- Get a table of the defined builds for the current specialization.
				local specBuilds = self.Available();

				-- Check to see if there is a table of builds, and that it contains a build with the name specified.
				if specBuilds and specBuilds[buildName] then
					-- The build exists so convert the talent string into a talent table and store it in the return value.
					returnValue = Conversion.ToTalents(specBuilds[buildName], specGroup);
				end
			end

			-- Return the stored talent table, or nil if the build name did not exist or the player has no specialization chosen.
			return returnValue;
		end,

		ChangesAllowed = function()
			local apiUnit = addon.API.Unit;
			local returnValue;

			-- Check to make sure the player is not in combat.
			if not apiUnit.InCombat() then
				-- Set the return value to true if the player is in a resting zone, or they have a valid preparation or tome buff.
				returnValue = addon.API.Blizzard.IsResting() or Buffs.PreperationActive() or Buffs.TomeActive();
			end

			-- Return whether talent changes are allowed.
			return returnValue or false;
		end,

		CurrentTalents = function()
			-- Get the current specialization group.
			local specGroup = addon.API.Blizzard.GetActiveSpecGroup();
			local returnValue;

			-- Check to see that the player has an active specialization.
			if specGroup then
				local talentID, name, texture, selected;
				-- Get the maximum row the player can select talents from.
				local numRows = addon.API.Blizzard.GetMaxTalentTier();

				-- Loop through the rows where the player could choose talents from.
				for row = 1, numRows do
					-- Loop through each talent in the row.
					for column = 1, 3 do
						-- Get the infomation about the talent being checked.
						talentID, name, texture, selected = addon.API.Blizzard.GetTalentInfo(row, column, specGroup);

						-- Check to see if the talent is selected.
						if selected then
							-- Check to see if the returnValue contains data.
							if not returnValue then
								-- It doesn't contain data so set it up as a new table.
								returnValue = {};
							end

							-- Store the talent name using the ID as the table key.
							returnValue[talentID] = name;
							-- We found a talent in this row, so lets just move on to the next row.
							break;
						end
					end
				end
			end

			-- Return the table of selected talents, or nil if the player doesn't have any talents selected.
			return returnValue;
		end,

		SelectedBuild = function()
			local self = addon.Features.TalentBuilds;

			-- Get the current specialization info and group.
			local specInfo = addon.API.General.GetSpecializationInfo();
			local specGroup = addon.API.Blizzard.GetActiveSpecGroup();
			local returnValue;

			-- Check to see that the player has an active specialization.
			if specGroup then
				local specBuilds = self.Available();

				-- Get the players current talents as a table.
				local currentTalents = self.CurrentTalents();
				-- Convert the talent table into a talent string.
				local talentString = Conversion.FromTalents(currentTalents, specGroup);

				-- Get the saved build name or the name of a build that matches current talents if a default build has not been selected.
				local buildName = addonTable.talentBuilds.CurrentBuild[specInfo.Name] or Builds.BuildName(talentString, specGroup);
				local buildTalentString = specBuilds[buildName];

				-- Set the return value data for the build.
				returnValue = {
					Name = addonTable.talentBuilds.CurrentBuild[specInfo.Name] or Builds.BuildName(talentString, specGroup),
					Talents = Conversion.ToTalents(buildTalentString, specGroup),
				};
			end

			-- Return the infomation found, or nil if the player does not have a specialization.
			return returnValue;
		end,

		Update = function()
			local self = addon.Features.TalentBuilds;
			local Tables = addon.Core.Tables;

			-- Get the current specialization info and group.
			local specInfo = addon.API.General.GetSpecializationInfo();
			local specGroup = addon.API.Blizzard.GetActiveSpecGroup();

			-- Check to see that the player has an active specialization.
			if specGroup then
				local specBuilds, presetBuilds, customBuilds = self.Available();

				-- Get the players current talents as a table.
				local currentTalents = self.CurrentTalents();
				-- Convert the talent table into a talent string.
				local talentString = Conversion.FromTalents(currentTalents, specGroup);

				-- Get the saved build name or the name of a build that matches current talents if a default build has not been selected.
				local buildName = addonTable.talentBuilds.CurrentBuild[specInfo.Name] or Builds.BuildName(talentString, specGroup);

				if buildName ~= addonTable.TalentBuilds.Custom then
					local buildTalentString = specBuilds[buildName];

					if not Conversion.Equate(talentString, buildTalentString, true) then
						if Tables.GetValue(addonTable.addonSettings, "autoLearnTalents", 0) == 1 then
							Builds.Update(buildTalentString);
						else
							local dialog = addon.API.Blizzard.StaticPopup_Show("LERH_UPDATE_BUILD", buildName);

							if dialog then
								dialog.data = buildName;
								dialog.data2 = buildTalentString;
							end
						end
					end
				end
			end
		end,
	};

	----------------------- Plugin Code Operation ----------------
	local moduleEvents = {
		playerLogin = function()
			local ElvUI = ElvUI;

			-- Check if ElvUI is Loaded
			useElvUI = addon.API.Blizzard.IsAddOnLoaded("ElvUI");

			-- Hook functions
			addon:RawHook("ToggleTalentFrame", Hooks.ToggleFrame, true);
			addon:RawHook("PanelTemplates_SetTab", Hooks.ChangeTab, true);

			if useElvUI then
				local engineElvUI = unpack(ElvUI);

				Skins = engineElvUI:GetModule("Skins");
			end
		end,

		changeSpecFinish = function()
			Events.PointsChanged();

			-- Send a callback that the build has been updated so minimap icon can change information.
			addon.System.Callbacks.Callback("updateMinimapIcon");
		end,

		rotationChanged = function()
			Events.PointsChanged();

			-- Send a callback that the build has been updated so minimap icon can change information.
			addon.System.Callbacks.Callback("updateMinimapIcon");
		end,
	};

	local tableEvents = {
		PLAYER_LEARN_TALENT_FAILED = Events.TalentFailed,
		PLAYER_LEVEL_UP = Events.PointsChanged,
		PLAYER_TALENT_UPDATE = Events.TalentUpdate,
		PLAYER_UPDATE_RESTING = Events.CheckAllowed,
		UNIT_AURA = Events.CheckAllowed,
	};

	-- These are only registered if the plugin is active as they control how what is called
	-- from LunaEclipse Rotation Helper
	addon.System.Callbacks.RegisterAddonCallbacks(moduleEvents);
	addon.System.Events.RegisterEvents(tableEvents, true);
end