-- Pulls back the Addon-Local Variables and store them locally.
local _, addonTable = ...;
local addon = _G["LunaEclipse_Rotation"];

-- Plugin information variables.
local pluginName = "Weak Auras";

-- Check whether the plugin is supposed to be active and that the required Weak Auras is loaded.
if addon and addon.API.Blizzard.IsAddOnLoaded("WeakAuras") and addon.System.Plugins.IsActive(pluginName) then
    -- Store local copies of any addon modules
    local styleName = "LunaEclipse";

    local LunaEclipse = addonTable.WeakAuras.Styles[styleName];
    local Icons = LunaEclipse.Data.Icons;
    local Texts = LunaEclipse.Data.Texts;

    local parentAura = {
        ["id"] = "LERH_LunaEclipse",
        ["regionType"] = "group",
    };

    local auraImports = {
        Icons.LERH_Defensive,
        Icons.LERH_Interrupt,
        Icons.LERH_OffGCD,
        Icons.LERH_AOE,
        Icons.LERH_CD,
        Icons.LERH_Bloodlust,
        Texts.LERH_Range,
        Texts.LERH_Enemies,
        Texts.LERH_TimeToDie,
        Texts.LERH_DamageAll,
        Texts.LERH_DamageMagic,
        Texts.LERH_DamagePhysical,
    };

    local weakAurasOptions = {
        RotationIcons = {
            args = {
                LERH_CD = {
                    type = "toggle",
                    name = "Cooldown",
                    desc = "Icon suggests any spells that trigger the global cooldown.\n\nSpells suggested will depend on your settings for the rotation.",
                    width = "full",
                    order = 10,
                },
                LERH_Defensive = {
                    type = "toggle",
                    name = "Defensive",
                    desc = "Icon is used to suggest defensive cooldowns when it is expected that you will take a lot of damage.",
                    width = "full",
                    order = 20,
                },
                LERH_Interrupt = {
                    type = "toggle",
                    name = "Interrupt",
                    desc = "Icon is used to suggest spells that can be used to interrupt spell casting.",
                    width = "full",
                    order = 30,
                },
                LERH_AOE = {
                    type = "toggle",
                    name = "Multi Target",
                    desc = "Icon displays the rotation to use based on the number of enemies you are fighting.",
                    width = "full",
                    order = 40,
                },
                LERH_OffGCD = {
                    type = "toggle",
                    name = "Off Global Cooldown",
                    desc = "Icon suggests any spells that don't trigger the global cooldown.\n\nSpells suggested will depend on your settings for the rotation.",
                    width = "full",
                    order = 50,
                },
                LERH_Bloodlust = {
                    type = "toggle",
                    name = "Bloodlust",
                    desc = "Icon displays when you have bloodlust and how long is remaining.\n\nThis is useful for lining up cooldowns.",
                    width = "full",
                    order = 60,
                },
                textDescription = {
                    type = "description",
                    name = "",
                    width = "full",
                    order = 70,
                },
            },
        },
        InfoTexts = {
            args = {
                LERH_Enemies = {
                    type = "toggle",
                    name = "Enemies",
                    desc = "This is a text display showing the number of enemies that the addon is currently tracking.",
                    width = "full",
                    order = 10,
                },
                LERH_DamageAll = {
                    type = "toggle",
                    name = "Incoming Damage",
                    desc = "This is a text display to display and estimated health percentage loss from any type of damage in the next 3 seconds.",
                    width = "full",
                    order = 20,
                },
                LERH_DamageMagic = {
                    type = "toggle",
                    name = "Incoming Magic Damage",
                    desc = "This is a text display to display and estimated health percentage loss from magic damage in the next 3 seconds.",
                    width = "full",
                    order = 30,
                },
                LERH_DamagePhysical = {
                    type = "toggle",
                    name = "Incoming Physical Damage",
                    desc = "This is a text display to display and estimated health percentage loss from physical damage in the next 3 seconds.",
                    width = "full",
                    order = 40,
                },
                LERH_Range = {
                    type = "toggle",
                    name = "Target Range",
                    desc = "This is a text display to display an estimated range to the target.",
                    width = "full",
                    order = 50,
                },
                LERH_TimeToDie = {
                    type = "toggle",
                    name = "Time to Die",
                    desc = "This is a text display of an estimated time to die for the current target.",
                    width = "full",
                    order = 60,
                },
                textDescription = {
                    type = "description",
                    name = "\nTypically the Text Auras are for debugging purposes, however some information provided may be useful while playing which is why they can be enabled.\n\nBy default these Weak Auras are disabled.",
                    width = "full",
                    order = 70,
                },
            },
        },
    };

    addon.Features.WeakAuras.RegisterStyle(styleName, weakAurasOptions, parentAura, auraImports);
end